/*
 * loader.c --		Target-side bootstrap loader for Armadillo.
 *
 * Copyright (C) 2001 Atmark Techno, Inc.
 * Copyright (C) 1999 Ben Williamson <benw@pobox.com>
 *
 * This program is loaded into SRAM in bootstrap mode, where it waits
 * for commands on UART1 to read and write memory, jump to code etc.
 * A design goal for this program is to be entirely independent of the
 * target board.  Anything with a Armadillo should be able to run
 * this code in bootstrap mode.  All the board specifics can be handled on
 * the host.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ioregs.h"
#include "cs8900.h"

#define DRAM_START	((unsigned *)0xc0000000)
#define DRAM_END	((unsigned *)0xe0000000)
#define STEP_BYTES	(256*1024)
#define STEP_WORDS	(STEP_BYTES / 4)
#define PATTERN		0x12345678


extern void flush_v3(void);
extern void flush_v4(void);


static unsigned char get_char(void)
{
	while (IO_SYSFLG1 & URXFE1);
	return IO_UARTDR1 & 0xff;
}

static void drain(void)
{
	while (IO_SYSFLG1 & UTXFF1);
}

static void put_char(unsigned char c)
{
	drain();
	IO_UARTDR1 = c;
}

static unsigned get_word(void)
{
	unsigned w;
	
	w = get_char();
	w |= get_char() << 8;
	w |= get_char() << 16;
	w |= get_char() << 24;
	return w;
}

static void put_word(unsigned w)
{
	put_char(w);
	put_char(w >> 8);
	put_char(w >> 16);
	put_char(w >> 24);
}

static void cs8900_reportmac()
{
	int i;

	/* report MAC address */
	for (i = 0; i < 6; i += 2) {
		unsigned short sh = get_reg(PP_IA+i);
		put_char(sh & 0xff);
		put_char(sh >> 8);
	}
}

/*
 * Initialize the onboard CS8900A Ethernet controller.
 */
static void cs8900_init()
{
	volatile int i;
	unsigned short 	us;

	/* check ID */
	if (get_reg(PP_ChipID) != 0x630e) {
		put_char('-');
		return;
	}

	/* reset */
	put_reg(PP_SelfCTL, PP_SelfCTL_Reset);
	for (i = 0; i < 3700000; i++)	/* approx. 200 ms */
		/* NOP */ ;
	i = 0x70000000;
	while ( (((us = get_reg(PP_SelfSTAT)) & PP_SelfSTAT_InitD) == 0) &&
			(i-- > 0) )
		/* NOP */ ;
	put_char(us & 0xff);
	put_char(us >> 8);

	/* enable receiver mode */
	put_reg(PP_RxCFG, PP_RxCFG_RxOK);
	put_reg(PP_RxCTL, PP_RxCTL_RxOK | PP_RxCTL_IA);
	put_reg(PP_LineCTL, PP_LineCTL_Rx);

	/* successful initialization */
	put_char('+');

	cs8900_reportmac();
}

/*
 * set a host provided MAC address for the onboard CS8900A Ethernet controller.
 */
static void cs8900_setmac()
{
	int i;
	unsigned short sh;

	/* set MAC address */
	for (i = 0; i < 6; i += 2) {
		sh = get_char();
		sh += get_char() << 8;
		put_reg(PP_IA+i, sh);
	}

	cs8900_reportmac();
}


/*
 * Get a data block via Ethernet.
 */
static unsigned char
cs8900_get_block(char *addr, int len)
{
	int i;
	unsigned short rxlen;
	unsigned char checksum = 0;

	while (!(get_reg(PP_RER) & PP_RER_RxOK));

	CS8900_RTDATA;		/* stat */
	rxlen = CS8900_RTDATA;	/* len */

	for (i=0; i<7; i++)
		CS8900_RTDATA;	/* 14 bytes for dst, src, and type */
	rxlen -= 14;

	/* assume even number of bytes */
	for (/* nothing */; rxlen > 0; rxlen -= 2) {
		unsigned short w = CS8900_RTDATA;
		*addr = w & 0xff;
		checksum += *addr++;
		*addr = w >> 8;
		checksum += *addr++;
	}
	return checksum;
}


/*
 * This is a destructive test for DRAM, which is safe because this program
 * runs entirely from internal SRAM.  First we detect the width of the DRAM,
 * which is reported by writing a single byte value of 16 or 32.  This is
 * followed by the compiled-in step size for scanning, which is 256kB.
 * 
 * We start at the end of the DRAM space and work backwards in 256kB steps,
 * writing the address of each block into its first word.  Then we scan
 * forward looking for blocks which contain their own address.  These are
 * the unique blocks of physical memory, which we report with word addresses.
 * The list is terminated with a zero address.
 */
static void detect_dram(void)
{
	volatile unsigned *p;

	IO_SYSCON2 &= ~DRAMSZ;		/* 32-bit wide */
	
	p = (unsigned *)DRAM_START;
	p[0] = PATTERN;
	p[1] = 0;			/* discharge data bus capacitance */
	if (p[0] != PATTERN) {
		IO_SYSCON2 |= DRAMSZ;	/* 16-bit wide it is then */
		put_char(16);
	} else {
		put_char(32);
	}
	put_word(STEP_BYTES);
	
	p = DRAM_END;
	while (p > DRAM_START) {
		p -= STEP_WORDS;
		*p = (unsigned)p;
	}
	/* p == DRAM_START */
	while (p < DRAM_END) {
		if (*p == (unsigned)p) {
			put_word((unsigned)p);
		}
		p+= STEP_WORDS;
	}
	put_word(0);
}


/*
 * Received:	start address
 *		length
 *
 * Transmitted:	data bytes
 *		checksum byte (sum of transmitted data)
 */
static void read_block(void)
{
	unsigned char checksum = 0;
	unsigned char *p = (unsigned char*) get_word();
	unsigned length = get_word();

	while (length--) {
		put_char(*p);
		checksum += *p++;
	}
	put_char(checksum);
}


/*
 * Received:	start address
 *		length
 *		data bytes...
 *
 * Transmitted:	checksum byte (sum of received data)
 */
static void write_block(void)
{
	unsigned char checksum = 0;
	unsigned char *p = (unsigned char*) get_word();
	unsigned length = get_word();

	while (length--) {
		*p = get_char();
		checksum += *p++;
	}
	put_char(checksum);
}


/*
 * Received:	start address
 *		length
 *		data bytes...
 *
 * Transmitted:	checksum byte (sum of received data)
 */
static void write_ethernet_block(void)
{
	unsigned char checksum;
	unsigned char *p = (unsigned char*) get_word();
	unsigned length = get_word();
	checksum = cs8900_get_block(p, length);
	put_char(checksum);
}


int
main(void)
{
	void (*code)(int r0, int r1, int r2, int r3);
	int r0, r1, r2, r3;
	volatile unsigned *w;
	volatile unsigned char *b;

	while (1) {
		switch (get_char()) {
		case '3':	/* Flush v3 MMU */
			flush_v3();
			break;
			
		case '4':	/* Flush v4 MMU */
			flush_v4();
			break;
			
		case 'a':	/* Ack please */
			put_char('+');	/* we're fine, thanks */
			break;
			
		case 'c':	/* Call (address, r0, r1, r2, r3) */
			code = (void *)get_word();
			r0 = get_word();
			r1 = get_word();
			r2 = get_word();
			r3 = get_word();
			drain();
			code(r0, r1, r2, r3);
			break;
		
		case 'd':	/* Detect DRAM */
			detect_dram();
			break;
		
		case 'e':	/* Initialize ethernet */
			cs8900_init();
			break;

		case 'M':	/* set MAC address */
			cs8900_setmac();
			break;

		case 'g':	/* Get single byte (address) */
			put_char(*(unsigned char *)get_word());
			break;
		
		case 's':	/* Set single byte (address, data) */
			b = (unsigned char *)get_word();
			*b = get_char();
			break;
		
		case 'r':	/* Read single word (address) */
			put_word(*(unsigned *)get_word());
			break;
		
		case 'w':	/* Write single word (address, data) */
			w = (unsigned *)get_word();
			*w = get_word();
			break;
		
		case 'R':	/* Read block */
			read_block();
			break;
		
		case 'W':	/* Write block */
			write_block();
			break;

		case 'E':	/* Write ethernet block */
			write_ethernet_block();
			break;
		
		default:
			put_char('?');
			break;
		}
	}
}

