/*
 * shoehorn.c --	Host-side bootstrap loader for Armadillo boards.
 * 
 * Copyright (C) 2001 Atmark Techno, Inc.
 * Copyright (c) 2000 Blue Mug, Inc.  All Rights Reserved.
 * Copyright (c) 1999 Ben Williamson <benw@pobox.com>
 *
 * This program loads a piece of code (see loader.c) into the target's
 * internal SRAM.  The loader then does our bidding to initialise and
 * detect hardware, load a kernel, ramdisk image, and boot.
 *
 * To add support for a new board type, add a new command-line option
 * that sets the hardware variable, and arrange for it to call
 * init_myboardname() instead of init_anvil().
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <getopt.h>
#include <errno.h>

#include "eth.h"
#include "ioregs.h"
#include "serial.h"
#include "util.h"
#include "cs8900.h"

#define _POSIX_SOURCE	1

#define MAKE_VERSION(ver) "shoehorn v" #ver " built " \
	__TIME__", "__DATE__ \
	" (CVS: $Id: shoehorn.c,v 1.10 2001/01/17 22:50:56 miket Exp $)";
static const char version[] 	= MAKE_VERSION(3.4-armadillo);


#define SRAM_SIZE	0x800		/* 2kB of SRAM for loader.bin */
#define START_CHAR	'<'
#define END_CHAR	'>'
#define MAX_FRAGS	100
#define DRAM_START	0xc0000000

#define MINORBITS	8
#define MKDEV(ma,mi)	(((ma) << MINORBITS) | (mi))
#define RAMDISK_MAJOR	1

/* architecture numbers are defined by the ARM kernel */
//#define ARCH_NUMBER_EDB7211	50
#define ARCH_NUMBER_ARMADILLO	83

#define PARAM_OFFSET	0x00020000	/* immediately after video RAM */
  #define PARAM_U1_OFFSET	0x00020000
  #define PARAM_U2_OFFSET	0x00020100
  #define PARAM_CMDLINE_OFFSET	0x00020500
  #define PARAM_END		0x00020900
  #define PARAM_SIZE		(256 + 1024 + 1024)

#define KERNEL_OFFSET	0x00028000	/* beginning of kernel image */
#define PAGE		0x1000

#define INITRD_START	0xc0400000

#define ETH_STEP	1024

static int ethernet = 0;
static int hardware = 0;
static int terminal = 0;
static int boot     = 0;

struct option options[] = {
	{ "armadillo",	0, &hardware,	'e' },
	{ "ethernet",	0, &ethernet,	1 },
	{ "boot",	0, &boot,	1 },
	{ "initrd",	1, 0,		'i' },
	{ "kernel",	1, 0,		'k' },
	{ "loader",	1, 0,		'l' },
	{ "netif",	1, 0,		'n' },
	{ "port",	1, 0,		'p' },
	{ "terminal",	0, &terminal,	1 },
	{ "version",	0, 0,		'v' },
	{ 0,		0, 0,		0 }
};

/* this magic is documented in the GNU cpp info files; we use
   it here to let a Makefile or build script specify the path
   to loader.bin. */
#ifdef LOADERPATH
#define loaderpath(s) str(s)
#else
#define loaderpath(s)
#endif
#define str(s) #s

static int arch_number	= -1;
static char *initrd	= "initrd";
static char *kernel	= "Image";
static char *loader	= loaderpath(LOADERPATH) "loader.bin";
static char *netif	= "eth0";
static char *port	= "/dev/ttyS0";

char *progname		= "UNKNOWN";

char kargs[256];
unsigned char remotemac[6];

struct fragment {
	unsigned int	start;
	unsigned int	size;
} frag_list[MAX_FRAGS], *frag;

/*
 * Print usage instructions and exit
 */
void
usage_and_exit(void)
{
	printf("Usage: %s [options] [kernel command line]\n"
	       "Available options (defaults):\n"
	       "        --boot\n"
	       "        --armadillo\n"
	       "        --ethernet\n"
	       "        --initrd (%s)\n"
	       "        --kernel (%s)\n"
	       "        --loader (%s)\n"
	       "        --netif (%s)\n"
	       "        --port (%s)\n"
	       "        --terminal\n"
	       "        --version\n",
	       progname, initrd, kernel, loader, netif, port);
	exit(1);
}


/*
 * Parse the command line options
 */
static void
parse_command_line(int argc, char **argv)
{
	int c;
	
	while (1) {
		c = getopt_long_only(argc, argv, "iklnp", options, NULL);
		if (c == -1) {
			break;
		}
		switch (c) {
		case 0:
			break;
		case 'i':
			initrd = optarg;
			break;
		case 'k':
			kernel = optarg;
			break;
		case 'l':
			loader = optarg;
			break;
		case 'n':
			netif = optarg;
			break;
		case 'p':
			port = optarg;
			break;
		case 'v':
			puts(version);
			exit(0);
		default:
			usage_and_exit();
		}
	}
	if (hardware == 0) {
		fprintf(stderr, "Need to specify hardware type (hint: --armadiloo)\n");
		usage_and_exit();
	}
}

/*
 * concatenating what's left into kargs, but after root privs are dropped
 */
static void
build_kargs(int argc, char **argv)
{
	int remspace;
	int prependspace = 0;	/* 0 on first arg, else 1 */

	memset(kargs, 0, sizeof kargs);
	while (optind < argc) {
		remspace = sizeof(kargs) - strlen(kargs) - 1;
		/* add if it fits, exit if too large */
		if (remspace >= strlen(argv[optind]) + prependspace) {
			/* will fit */
			if (prependspace)	strcat(kargs, " ");
			strcat(kargs, argv[optind++]);
		} else {
			printf("Kernel options too long\n");
			exit(1);
		}
		prependspace = 1;
	}
	kargs[sizeof(kargs) - 1] = '\0';
}


void target_write_ethernet(unsigned addr, const char *buf,
			   unsigned size, unsigned progress)
{
	unsigned char frame [2048];

	while (size > 0) {
		unsigned char checksum = 0;
		unsigned char targetsum;
		unsigned step;
		int response, tries = 5;

		step = min(size, ETH_STEP);
		put_char('E');
		put_word(addr);
		put_word(step);

		/* ethernet frame format:
		   6-byte dest, 6-byte src, 2-byte dummy type and <=
		   1024-byte data data size must be an even number */

		memset(frame, 0, sizeof frame);
		memcpy(frame, remotemac, 6);
		*(unsigned short*)(frame + 12) = 0xabba;
		memcpy(frame + 14, buf, step);

		do {
			eth_write(frame, step + 14);
			targetsum = response = get_char_timeout(500);
		} while ((response < 0) && --tries);

		if (!tries) {
			printf("\nEthernet: no answer\n");
			exit(1);
		}

		addr += step;
		size -= step;
		progress += step;
		while (step-- > 0) {
			checksum += *buf++;
		}
		if (checksum != targetsum) {
			fprintf(stderr, "\nEthernet checksum error\n");
			exit(1);
		}
		printf("0x%08x\r", progress);
		fflush(NULL);
	}
}

void target_write(unsigned addr, const char *buf,
		  unsigned size, unsigned progress)
{
	/* XXX this is kind of nasty */
	if (ethernet)
		target_write_ethernet(addr, buf, size, progress);
	else
		target_write_block(addr, buf, size, progress);
}

/*
 * Write an object to the target, spanning multiple DRAM fragments
 * Returns final address
 */
unsigned int
target_write_fragmented(unsigned int addr, char *buf, unsigned int size)
{
	struct fragment *frag;
	unsigned int frag_end = 0;
	int progress = 0;
	
	for (frag = &frag_list[1]; frag->size != 0; frag++) {
		frag_end = frag->start + frag->size;
		if ((frag->start <= addr) && (addr < frag_end)) {
			break;
		}
	}
	while (size > 0) {
		int step = min(size, frag_end - addr);
		if (step <= 0) {
			printf("Insufficient DRAM space\n");
			exit(1);
		}
		target_write(addr, buf, step, progress);
		addr += step;
		buf += step;
		progress += step;
		size -= step;
		if (size > 0) {
			frag++;
			addr = frag->start;
			frag_end = frag->start + frag->size;
		}
	}
	if (addr == frag_end) {
		addr = frag[1].start;
	}
	return addr;
}


void
ping(void)
{
	char c;

	printf("Pinging loader\n");
	put_char('a');
	c = get_char();
	if (c != '+') {
		printf("Expected acknowledge '+' but got '%c'\n", c);
		exit(1);
	}
}

/*
 * This definition of param_struct is copied from include/asm-arm/setup.h
 * in Linux 2.4.0-test6-rmk5.  Changes made after that kernel should
 * hopefully remain backwards compatible.
 *
 * Note that these will have to be changed to e.g. 'u32' to work on a 64-
 * bit development workstation.
 */
#define COMMAND_LINE_SIZE 1024

struct param_struct {
    union {
	struct {
	    unsigned long page_size;		/*  0 */
	    unsigned long nr_pages;		/*  4 */
	    unsigned long ramdisk_size;		/*  8 */
	    unsigned long flags;		/* 12 */
#define FLAG_READONLY	1
#define FLAG_RDLOAD	4
#define FLAG_RDPROMPT	8
	    unsigned long rootdev;		/* 16 */
	    unsigned long video_num_cols;	/* 20 */
	    unsigned long video_num_rows;	/* 24 */
	    unsigned long video_x;		/* 28 */
	    unsigned long video_y;		/* 32 */
	    unsigned long memc_control_reg;	/* 36 */
	    unsigned char sounddefault;		/* 40 */
	    unsigned char adfsdrives;		/* 41 */
	    unsigned char bytes_per_char_h;	/* 42 */
	    unsigned char bytes_per_char_v;	/* 43 */
	    unsigned long pages_in_bank[4];	/* 44 */
	    unsigned long pages_in_vram;	/* 60 */
	    unsigned long initrd_start;		/* 64 */
	    unsigned long initrd_size;		/* 68 */
	    unsigned long rd_start;		/* 72 */
	    unsigned long system_rev;		/* 76 */
	    unsigned long system_serial_low;	/* 80 */
	    unsigned long system_serial_high;	/* 84 */
	    unsigned long mem_fclk_21285;       /* 88 */ 
	} s;
	char unused[256];
    } u1;
    union {
	char paths[8][128];
	struct {
	    unsigned long magic;
	    char n[1024 - sizeof(unsigned long)];
	} s;
    } u2;
    char commandline[COMMAND_LINE_SIZE];
};


void
target_write_params(unsigned long initrd_start, unsigned long initrd_size)
{
	struct param_struct ps;

	/* sanity checking to keep updates from breaking anything */
	assert(PARAM_SIZE == PARAM_END - PARAM_OFFSET);
	assert(sizeof (struct param_struct) <= PARAM_SIZE);

	/* XXX these need to be generalized for different HW setups */
	/* XXX they are also endian specific */
	memset(&ps, 0, sizeof ps);
	/* printf("- page_size: %d\n", PAGE); */
	ps.u1.s.page_size = PAGE;
	printf("- nr_pages (all banks): %d\n", 4096);
	ps.u1.s.nr_pages = 4096;
	/* XXX */
	ps.u1.s.flags = FLAG_READONLY | FLAG_RDLOAD | FLAG_RDPROMPT;
	printf("- rootdev: (RAMDISK_MAJOR, 0)\n");
	ps.u1.s.rootdev = MKDEV(RAMDISK_MAJOR, 0);
	printf("- pages_in_bank[0]: %d\n", 2048);
	ps.u1.s.pages_in_bank[0] = 2048;
	printf("- pages_in_bank[1]: %d\n", 2048);
	ps.u1.s.pages_in_bank[1] = 2048;
	printf("- initrd_start: 0x%lx\n", initrd_start);
	ps.u1.s.initrd_start = initrd_start;
	printf("- initrd_size: 0x%lx\n", initrd_size);
	ps.u1.s.initrd_size = initrd_size;
	printf("- ramdisk_size: 0x%lx\n", initrd_size);
	ps.u1.s.ramdisk_size = initrd_size;

	/* XXX these writes overlap */
	print_size(DRAM_START + PARAM_OFFSET, PARAM_SIZE);
	target_write_fragmented(DRAM_START + PARAM_OFFSET,
		(char*) &ps, PARAM_SIZE);
	target_write_fragmented(DRAM_START + PARAM_CMDLINE_OFFSET,
		kargs, sizeof(kargs));
}

	
void
init_armadillo(void)
{
	printf("- flushing cache/TLB\n");
	put_char('4');

	/* PADDR = 0xfc */
	/* PBDDR = 0x00 */
	/* PDDDR = 0x18 */
	/* PEDDR = 0xff */
	target_write_word(IO(PADDR), 0xfc);
	target_write_word(IO(PBDDR), 0x00);
	target_write_word(IO(PDDDR), 0x18);
	target_write_word(IO(PEDDR), 0xff);

	printf("- 73MHz core clock\n");
	/* IO_SYSCON3 = (IO_SYSCON3 & ~CLKCTL) | CLKCTL_73 */
	target_write_word(IO(SYSCON3),
		(target_read_word(IO(SYSCON3)) & ~CLKCTL) | CLKCTL_73);
	
#if 0
	printf("- 64kHz DRAM refresh\n");
	/* IO_DRFPR = 0x81 */
	target_write_word(IO(DRFPR), 0x81);
#else
	//target_write_word(IO(SDCONF), SDR_CAS_LAT_2 | SDR_CAP_128
			  //| SDR_WIDTH_16 | SDR_CLK_PERM | SDR_ENABLE);
	//target_write_word(IO(SDRFPR), 0x100);
	target_write_word(IO(0x2300), 0x2 | 0x40 | 0x100 | 0x0 | 0x400);
	target_write_word(IO(0x2340), 0x100);
#endif

	printf("- enabling UART2\n");
	/* IO_PDDR = 0x10 */
	/* XXX is this right for Armadillo? */
	target_write_byte(IO(PDDR), 0x10); /* enable AuxPwrGate transceiver */
	/* IO_SYSCON2 |= UART2EN */
	target_write_word(IO(SYSCON2),
		target_read_word(IO(SYSCON2)) | UART2EN);

	printf("- Activate LED flasher\n");
	/* IO_LEDFLSH = 0x40 */
	target_write_byte(IO(LEDFLSH), 0x40);

	printf("- Setting up flash at CS0\n");
	/* IO_MEMCFG1 = (IO_MEMCFG1 & 0xffffff00) | 0x00000094 */
	target_write_word(IO(MEMCFG1),
		(target_read_word(IO(MEMCFG1)) & 0xffffff00) | 0x00000094);

	printf("- Setting up Ethernet at CS2\n");
	/* IO_MEMCFG1 = (IO_MEMCFG1 & 0xff00ffff) | 0x00840000 */
	target_write_word(IO(MEMCFG1),
		(target_read_word(IO(MEMCFG1)) & 0xff00ffff) | 0x00840000);

	printf("- Setting up Miscellaneous(8bit-width) at CS3\n");
	/* IO_MEMCFG1 = (IO_MEMCFG1 & 0x00ffffff) | 0x9b000000 */
	target_write_word(IO(MEMCFG1),
		(target_read_word(IO(MEMCFG1)) & 0x00ffffff) | 0x9b000000);

	printf("- Setting up Miscellaneous(16bit-width) at CS4\n");
	/* IO_MEMCFG2 = (IO_MEMCFG2 & 0xffffffff) | 0x00000098 */
	target_write_word(IO(MEMCFG2),
		(target_read_word(IO(MEMCFG2)) & 0xffffff00) | 0x00000098);

	/* 0x3000000e = 0x00 */
	target_write_byte(0x3000000e, 0x00);

	printf("Switching to 115200 baud\n");
	/* IO_UBRLCR1 = IO_UBRLCR1 & ~BRDIV | BR_115200 */
	target_write_word(IO(UBRLCR1),
		(target_read_word(IO(UBRLCR1)) & ~BRDIV) | BR_115200);
	serial_baud(B115200);
}

#define UBRLCR2	0x14c0	/* UART Bit Rate and Line Control register --------- */

void
post_armadillo(void)
{
	printf("Setting up UART2\n");
	target_write_word(IO(UBRLCR2), (WL_8<<WRDLEN_SHIFT) | FIFOEN | PRTEN | BR_115200);

#if 0
	printf("Switching back to 9600 baud\n");
	/* IO_UBRLCR1 = IO_UBRLCR1 & ~BRDIV | BR9600 */
	target_write_word(IO(UBRLCR1),
		(target_read_word(IO(UBRLCR1)) & ~BRDIV) | BR_9600);
	serial_baud(B9600);
#endif
}


void
detect_dram(void)
{
	unsigned int start, size, total_size;

	put_char('d');
	printf("- %d bits wide\n", get_char());
	frag_list[0].start = 0;	/* item 0 is a dummy, list starts at 1 */
	frag_list[0].size = 0;
	total_size = 0;
	frag = frag_list;
	size = get_word();
	while (1) {
		start = get_word();
		if (start == 0) {
			break;
		}
		if (start == frag->start + frag->size) {
			frag->size += size;
		} else {
			frag++;
			if (frag >= &frag_list[MAX_FRAGS-1]) {
				printf("Too many DRAM fragments\n");
				exit(1);
			}
			frag->start = start;
			frag->size = size;
		}
		total_size += size;
	}
	frag++;
	frag->start = frag->size = 0;
	for (frag = &frag_list[1]; frag->size != 0; frag++)
		print_size(frag->start, frag->size);
	printf("Total DRAM: %dkB\n", total_size / 1024);
}


void perror_usage_exit(const char *s)
{
	fprintf(stderr, "%s: ", progname);
	if (errno > sys_nerr - 1) {
		fprintf(stderr, "%s: unknown error # %d\n", s, errno);
	} else {
		fprintf(stderr, "%s: %s\n", s, sys_errlist[errno]);
	}
	usage_and_exit();
}


/*
 * Send loader.bin to the target
 * Initialise some target registers, for Anvil hardware
 * Detect what DRAM is present, building a list of fragments
 * Load the kernel image and initrd into available fragments
 * Write a kernel parameter block with command-line arguments
 * Jump to the kernel entry point
 */
int
main(int argc, char **argv)
{
	unsigned char *initrd_buf, *kernel_buf, *loader_buf;
	unsigned kernel_size, initrd_size, loader_size;
	unsigned long kernel_end, initrd_start = INITRD_START, size;
	int i;
	uid_t ruid, euid, suid;
	int getresuid(uid_t *, uid_t *, uid_t *);	/* linux only????? */
	
	progname = strrchr(argv[0], '/');
	if (progname == NULL) {
		progname = argv[0];
	} else {
		progname++;
	}
	parse_command_line(argc, argv);

	/* make output to stdout immediately visible */
	setbuf(stdout, NULL);

	/* initialize Ethernet */
	if (ethernet) {
		printf("Initializing local network interface\n");
		eth_open(netif);
	}

	/* drop root privs if setuid root */
	if ( (geteuid() == 0) && (getuid() != 0) ) {
		printf("Dropping root privileges\n");
		if (setuid(getuid()) != 0) {
			fprintf(stderr, "%s: error in droping root-privs\n",
								progname);
			exit(EXIT_FAILURE);
		}
		/* FIXME: make it portable to systems w/o saved uid */
		/* check if root privs are really dropped */
		if (getresuid(&ruid, &euid, &suid) != 0) {
			fprintf(stderr, "%s: could not check if root-privs "
						"were dropped\n", progname);
			exit(EXIT_FAILURE);
		}
		if ( (ruid == 0) || (euid == 0) || (suid == 0) ) {
			fprintf(stderr, "%s: not all root-privs were dropped\n"
					"\truid: %d, euid: %d, suid: %d\n",
				       	progname, ruid, euid, suid);
			exit(EXIT_FAILURE);
		}
	}

	/* fill in kargs *after* dropping privileges */
	build_kargs(argc, argv);

	/* slurp files into buffers */
	loader_size = SRAM_SIZE;  /* must allocate at least SRAM_SIZE bytes */
	read_file(loader, &loader_buf, &loader_size);
	kernel_size = 0;
	read_file(kernel, &kernel_buf, &kernel_size);
	initrd_size = 0;
	read_file(initrd, &initrd_buf, &initrd_size);

	/* make sure loader isn't too big */
	if (loader_size > SRAM_SIZE) {
		fprintf(stderr, "%s: loader too large (limit %d bytes)\n",
			progname, SRAM_SIZE);
		exit(1);
	}
	if (loader_size > SRAM_SIZE - 0x100)
		fprintf(stderr, "%s: warning: loader stack might clobber code\n",
				progname);

	/* open serial port and start talking to hardware */
	serial_open(port);
	printf("Waiting for target - press Wakeup now.\n");
	while (get_char() != START_CHAR)
	  ;
	printf("Writing SRAM loader...\n");
	put_block(loader_buf, SRAM_SIZE);
	if (get_char() != END_CHAR) {
		printf("Expected end character '%c'\n", END_CHAR);
		exit(1);
	}
	ping();
	
	switch(hardware) {
	case 'e':
		arch_number = ARCH_NUMBER_ARMADILLO;
		printf("Initialising Armadillo hardware:\n");
		init_armadillo();
		break;
	default:
		printf("Internal error - invalid hardware value\n");
		exit(1);
	}
	assert(arch_number > 0);
	ping();

	if (ethernet) {
		unsigned	allff;
		unsigned	allzero;
		unsigned short	status;

		printf("Initializing remote Ethernet\n");
		put_char('e');
		status = get_char();
		status += get_char() << 8;
		printf("cs8900 status: %04x (", status);
		if (status & PP_SelfSTAT_EEPROM) {
			printf("EEPROM present:");
			if (status & PP_SelfSTAT_EEPROM_OK)
				printf(" OK, ");
			printf(" size = ");
			if (status & PP_SelfSTAT_EEsize) {
				printf("64");
			} else {
				printf("128/256");
			}
			printf(" words");
		}
		printf(")\n");
		if (get_char() != '+') {
			fprintf(stderr, "%s: Ethernet initialization error\n",
					progname);
			exit(1);
		}
		printf("MAC address of target is ");
		allff = 1;
		allzero = 1;
		for (i=0; i<6; i++) {
			remotemac[i] = get_char();
			if (remotemac[i] != 0xff) {
				allff = 0;
			}
		       	if (remotemac[i] != 0x00) {
				allzero = 0;
			}
			putchar('.');
		}
		putchar(' ');
		allff |= allzero;
		if (allff) {
			printf("uninitialized; setting.\n");
			put_char('M');
			/* set 12:34:56:78:9a:bc as MAC address */
			for (i=0; i<6; i++) {
				put_char(0x12 + i * 0x22);
			}
			printf("Now MAC address of device is ");
		}
		for (i=0; i<6; i++) {
			remotemac[i] = allff ? get_char() : remotemac[i];
			printf("%02X%c", remotemac[i], i==5 ? '\n' : ':');
		}
	}

	printf("Detecting DRAM\n");
	detect_dram();
	
	printf("Loading %s:\n", kernel);
	print_size(DRAM_START + KERNEL_OFFSET, kernel_size);
	kernel_end = target_write_fragmented(DRAM_START + KERNEL_OFFSET,
		kernel_buf, kernel_size);
	free(kernel_buf);

	/* Find a start address for initrd.  We put it
	   at the highest possible page-aligned address. */
	size = (initrd_size + PAGE - 1) & ~(PAGE - 1);
	for (frag = &frag_list[1]; frag->size != 0; frag++);
	frag--;
	while ((size > 0) && (frag > frag_list)) {
		int step = min(size, frag->size);
		initrd_start = frag->start + frag->size - step;
		size -= step;
		frag--;
	}
	/* XXX miket: trying this out */
	printf("initrd_start is %lx\n", initrd_start);
	initrd_start = INITRD_START;
	printf("Moving initrd_start to %lx\n", initrd_start);
	/* end XXX area */
	if (initrd_start < kernel_end) {
		printf("Not enough space for initrd\n");
		exit(1);
	}

	printf("Loading %s:\n", initrd);
	print_size(initrd_start, initrd_size);
	target_write_fragmented(initrd_start, initrd_buf, initrd_size);
	free(initrd_buf);
	
	printf("Writing parameter area\n");
	target_write_params(initrd_start, initrd_size);
	
	switch(hardware) {
	case 'e':
		post_armadillo();
		break;
	default:
		fprintf(stderr, "%s: Internal error - invalid hardware value\n",
				progname);
		exit(1);
	}
	ping();

	if (ethernet)
		eth_close();
	
	if (boot) {
		printf("Starting kernel\n");
		put_char('c');
		put_word(DRAM_START + KERNEL_OFFSET);
		put_word(0);	/* sanity check */
		put_word(arch_number);
		put_word(0);
		put_word(0);
	}

	/* I've found this bit useful for debugging: If the kernel doesn't
	   seem to boot, run "shoehorn --terminal" and plug head-armv.S
	   and init/main.c full with things like "IO_UARTDR1 = '!';" or
	   similar.  Then you'll find out how far the setup code gets. */
	if (terminal)
		serial_terminal();

	serial_close();
	return 0;
}

